package cn.com.qmth.examcloud.web.support;

import cn.com.qmth.examcloud.api.commons.exchange.BaseResponse;
import cn.com.qmth.examcloud.api.commons.exchange.JsonSerializable;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLog;
import cn.com.qmth.examcloud.commons.logging.ExamCloudLogFactory;
import cn.com.qmth.examcloud.commons.util.JsonMapper;
import cn.com.qmth.examcloud.commons.util.ObjectUtil;
import cn.com.qmth.examcloud.commons.util.ThreadLocalUtil;
import cn.com.qmth.examcloud.web.config.LogProperties;
import cn.com.qmth.examcloud.web.enums.HttpServletRequestAttribute;
import cn.com.qmth.examcloud.web.redis.RedisClient;
import org.apache.commons.lang3.ArrayUtils;
import org.aspectj.lang.ProceedingJoinPoint;
import org.aspectj.lang.annotation.Around;
import org.aspectj.lang.annotation.Aspect;
import org.aspectj.lang.annotation.Pointcut;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.ResponseEntity;
import org.springframework.stereotype.Component;
import org.springframework.web.multipart.MultipartFile;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Collection;
import java.util.Map;

/**
 * spring mvc controller aspect.
 *
 * @author WANGWEI
 * @date 2019年1月30日
 * @Copyright (c) 2018-2020 WANGWEI [QQ:522080330] All Rights Reserved.
 */
@Component
@Aspect
public class ControllerAspect {

    private static final ExamCloudLog DEBUG_LOG = ExamCloudLogFactory.getLog(ControllerAspect.class);

    private static final ExamCloudLog INTERFACE_LOG = ExamCloudLogFactory.getLog("INTERFACE_LOGGER");

    @Autowired(required = false)
    HttpMethodProcessor httpMethodProcessor;

    @Autowired
    LogProperties logProperties;

    @Autowired(required = false)
    RedisClient redisClient;

    /**
     * 构造函数
     */
    public ControllerAspect() {
        super();
        DEBUG_LOG.info("ControllerAspect init...");
    }

    private static String[] excludeFields = new String[]{"password", ".*Password"};

    /**
     * handlerMethods
     *
     * @author WANGWEI
     */
    @Pointcut("@annotation(org.springframework.web.bind.annotation.RequestMapping) "
            + "|| @annotation(org.springframework.web.bind.annotation.GetMapping) "
            + "|| @annotation(org.springframework.web.bind.annotation.PostMapping)"
            + "|| @annotation(org.springframework.web.bind.annotation.DeleteMapping)"
            + "|| @annotation(org.springframework.web.bind.annotation.PutMapping)")
    public void handlerMethods() {

    }

    /**
     * 接口处理
     *
     * @param joinPoint
     * @return
     * @author WANG WEI
     */
    @Around("handlerMethods()")
    public Object doAroundWebRequests(ProceedingJoinPoint joinPoint) {
        // 获取request对象
        HttpServletRequest request = ServletUtil.getRequest();
        String path = request.getServletPath();
        try {
            if ("/".equals(request.getServletPath())) {
                return joinPoint.proceed();
            }
        } catch (Throwable e) {
            INTERFACE_LOG.error(e.getMessage());
            throw new RuntimeException(e);
        }

        long startTime = System.currentTimeMillis();
        String method = request.getMethod();
        Object[] args = joinPoint.getArgs();

        if (INTERFACE_LOG.isInfoEnabled()) {
            StringBuilder params = new StringBuilder();

            if (args != null) {
                if (args.length == 0) {
                    Map<String, String[]> paramMaps = request.getParameterMap();
                    for (Map.Entry<String, String[]> entry : paramMaps.entrySet()) {
                        String name = entry.getKey();
                        String[] values = entry.getValue();
                        params.append("&").append(name).append("=");
                        if (ArrayUtils.isNotEmpty(values)) {
                            if (name.toLowerCase().indexOf(excludeFields[0]) >= 0) {
                                params.append("***");
                            } else {
                                params.append(values[0]);
                            }
                        }
                    }
                } else {
                    for (int i = 0; i < args.length; i++) {
                        Object curArg = args[i];
                        if (curArg == null) {
                            params.append("null");
                        } else if (ObjectUtil.isBaseDataType(curArg.getClass())) {
                            params.append(curArg);
                        } else {
                            String content = this.parseContent(curArg);
                            if (content == null) {
                                content = "...";
                            } else {
                                if (content.length() > logProperties.getResponseLogJsonMaxSize()) {
                                    content = "content too large...";
                                }
                            }
                            params.append(content);
                        }
                        if (i < args.length - 1) {
                            params.append(" | ");
                        }
                    }
                }
            }

            INTERFACE_LOG.info(String.format("[ControllerAspect][request][%s] - %s, params = %s", method, path, params.toString()));
        }

        Object ret;
        try {
            if (null != httpMethodProcessor) {
                httpMethodProcessor.beforeMethod(request, args);
            }

            // 执行
            ret = joinPoint.proceed();

            request.setAttribute(HttpServletRequestAttribute.$_EXCEPTION_HAPPENED.name(), new Boolean(false));
        } catch (Throwable e) {
            request.setAttribute(HttpServletRequestAttribute.$_EXCEPTION_HAPPENED.name(), new Boolean(true));

            if (null != httpMethodProcessor) {
                try {
                    httpMethodProcessor.onException(request, args, e);
                } catch (Exception ex) {
                    INTERFACE_LOG.error("[ControllerAspect] processor1 - " + ex.getMessage(), ex);
                }
            }

            INTERFACE_LOG.error(String.format("[ControllerAspect][execute fail][%s] - %s, cost %sms, err is %s"
                    , method, path, System.currentTimeMillis() - startTime, e.getMessage()));
            throw new RuntimeException(e);
        }

        if (null != httpMethodProcessor) {
            try {
                httpMethodProcessor.onSuccess(request, args, ret);
            } catch (Exception ex) {
                INTERFACE_LOG.error("[ControllerAspect] processor2 - " + ex.getMessage(), ex);
            }
        }

        if (null != ret && BaseResponse.class.isAssignableFrom(ret.getClass())) {
            BaseResponse baseResponse = (BaseResponse) ret;
            baseResponse.setCost(System.currentTimeMillis() - startTime);
        }

        if (INTERFACE_LOG.isDebugEnabled() && logProperties.isNormalResponseLogEnable()) {
            if (ret == null) {
                INTERFACE_LOG.debug("[ControllerAspect] status = 200, responseMsg = void");
            } else if (ret instanceof ResponseEntity) {
                ResponseEntity<?> re = (ResponseEntity<?>) ret;
                Object body = re.getBody();

                String content = null;
                if (body != null) {
                    boolean jsonSerializable = false;
                    if (body instanceof Collection) {
                        jsonSerializable = true;
                    } else if (body instanceof Map) {
                        jsonSerializable = true;
                    } else if (body instanceof String) {
                        jsonSerializable = true;
                    } else if (body instanceof JsonSerializable) {
                        jsonSerializable = true;
                    }
                    if (jsonSerializable) {
                        content = this.parseContent(body);
                    }
                }

                if (content == null) {
                    content = "...";
                } else {
                    if (content.length() > logProperties.getResponseLogJsonMaxSize()) {
                        content = "content too large...";
                    }
                }

                INTERFACE_LOG.debug(String.format("[ControllerAspect] status = %s, responseMsg = %s",
                        re.getStatusCodeValue(), content));
            } else {
                String content;
                if (ObjectUtil.isBaseDataType(ret.getClass())) {
                    content = ret.toString();
                } else {
                    content = this.parseContent(ret);
                }

                if (content == null) {
                    content = "...";
                } else {
                    if (content.length() > logProperties.getResponseLogJsonMaxSize()) {
                        content = "content too large...";
                    }
                }

                INTERFACE_LOG.debug(String.format("[ControllerAspect] status = 200, responseMsg = %s", content));
            }
        }

        HttpServletResponse response = ServletUtil.getResponse();
        response.setHeader("Trace-Id", ThreadLocalUtil.getTraceId());
        response.setHeader("cost", String.valueOf(System.currentTimeMillis() - startTime));

        INTERFACE_LOG.info(String.format("[ControllerAspect][response][%s] - %s, cost %sms", method, path, System.currentTimeMillis() - startTime));
        return ret;
    }

    private String parseContent(Object obj) {
        try {
            if (obj instanceof MultipartFile) {
                return ((MultipartFile) obj).getOriginalFilename();
            }

            // return new JsonMapper().toJson(obj);
        } catch (Exception e) {
            INTERFACE_LOG.warn("[ControllerAspect] parseContent " + e.getMessage());
        }

        return null;
    }

}
